//------------------------------------------------------------------------------
//Copyright (C) 2013 by Gusev Alexander
//e-mail: alphacity@rambler.ru
//---------------------------------------------------------------------------
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <vector>
#include <string.h>
#include <errno.h>
#include <signal.h>
//---------------------------------------------------------------------------
#include "platform.h"
#include "socket_engine.h"
#include "printf_ex.h"
#include "time_utils.h"
#include "proxy_utils.h"
#include "http_utils.h"
#include "thread_obj.h"
#include "get_app_name.h"
#define DEBUG
//#define CS_ENABLED
#define FIREBIRD
//---------------------------------------------------------------------------
#ifdef FIREBIRD

#include <ibase.h>

#else

#ifdef IS_LINUX

#include <mysql/mysql.h>

#else

#include <mysql.h>

#endif

#endif

#ifdef IS_LINUX

#include <sys/time.h>

#endif
//---------------------------------------------------------------------------
static std::string btx_marker = "^"; //$
static std::string data_separator = ":"; //,
static std::string cs_marker = "&"; //*
static std::string etx_marker = "|"; //\r

#ifdef FIREBIRD

static ISC_STATUS_ARRAY status;
static isc_db_handle db_h = 0;

static std::string firebird_error(ISC_STATUS *status)
{
    std::string res;
    char buffer[512];
    int idx = 0;
    char idx_str[16];
    while (isc_interprete(buffer, &status))
    {
        sprintf(idx_str, "%i ", idx++);
        res.append(idx_str);
        res.append(buffer);
        res.append("\n");
    }
    return res;
}

static bool firebird_exec(char *query)
{
    bool res = false;

    isc_tr_handle tr_h = 0;

    while (true) {

    isc_start_transaction(status, &tr_h, 1, &db_h, 0, 0);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_start_transaction error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        tr_h = 0;
        break;
    }

    isc_dsql_execute_immediate(status, &db_h, &tr_h, 0, query, 1, 0);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_execute_immediate error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        break;
    }

    isc_commit_transaction(status, &tr_h);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_commit_transaction error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        tr_h = 0;
        break;
    }
    tr_h = 0;

    res = true;

    break;} //while (true) {

    if (tr_h)
    {
        isc_rollback_transaction(status, &tr_h);
        if (status[0] == 1 && status[1])
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s isc_rollback_transaction error:\n"
                "%s",
                system_time_str().c_str(),
                firebird_error(status).c_str()
            );
            #endif
        }
    }

    return res;
}

#else

static MYSQL mysql;

#endif

static std::string db_address = "localhost";
static unsigned short db_port = 0;
static std::string db_login = "admin";
static std::string db_password = "12345";
static std::string db_name = "db";

static int good_queries_count;
static int bad_queries_count;

static std::map<std::string, std::string> params;
//---------------------------------------------------------------------------
static unsigned short get_cs(unsigned char *beg, unsigned char *end)
{
    unsigned short res = 0;
    unsigned short c;
    while (beg != end)
    {
        c = *beg++;
        if (c != '$') res ^= c;
    }
    return res;
}
//---------------------------------------------------------------------------
static bool hex1_to_ui2(short ch, unsigned short * const res)
{
    bool good = false;
    *res = 0;

    ch -= 48;
    if (ch >= 0)
    {
        if (ch > 9) ch -= 7;
        if (ch <= 15)
        {
            good = true;
            *res = ch;
        }
    }

    return good;
}
//---------------------------------------------------------------------------
static bool get_machine_id(const std::string &machine_id_1, std::string * const machine_id_2, const bool get_virtual_machine_id)
{
    bool res = false;
    machine_id_2->clear();

    #ifdef FIREBIRD
////////////////////////////////////////////////////////////////////////////////
    isc_tr_handle tr_h = 0;

    isc_stmt_handle st_h = 0;

    XSQLDA *sqlda = (XSQLDA *)malloc(XSQLDA_LENGTH(1));
    sqlda->sqln = 1;
    sqlda->version = 1;

    char *data0 = 0;

    while (true) {

    isc_start_transaction(status, &tr_h, 1, &db_h, 0, 0);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_start_transaction error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        tr_h = 0;
        break;
    }

    isc_dsql_allocate_statement(status, &db_h, &st_h);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_allocate_statement error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        st_h = 0;
        break;
    }

    isc_dsql_prepare(status, &tr_h, &st_h, 0, (char *)(std::string() +
        "SELECT `" + (get_virtual_machine_id? "id" : "imei") + "` FROM `machines` "
        "WHERE `" + (get_virtual_machine_id? "imei" : "id") + "` = " +
        (get_virtual_machine_id? "'" : "") + machine_id_1 + (get_virtual_machine_id? "'" : "")).c_str(), 1, sqlda);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_prepare error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        break;
    }

    short ind0 = 0;

    //   :
    //2 -  vary_length
    //sqllen - .  vary_string
    //1 -  

    if (get_virtual_machine_id)
    {
        sqlda->sqlvar[0].sqldata = data0 = new char[sizeof(long)];
        sqlda->sqlvar[0].sqltype = SQL_LONG + 1;
        sqlda->sqlvar[0].sqlind  = &ind0;
    }
    else
    {
        sqlda->sqlvar[0].sqldata = data0 = new char[2 + sqlda->sqlvar[0].sqllen + 1];
        sqlda->sqlvar[0].sqltype = SQL_VARYING + 1;
        sqlda->sqlvar[0].sqlind  = &ind0;
    }

    isc_dsql_execute(status, &tr_h, &st_h, 1, 0);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_execute error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        break;
    }

    ISC_STATUS retcode;
    while (true)
    {
        retcode = isc_dsql_fetch(status, &st_h, 1, sqlda);
        if (retcode) break;

        if (get_virtual_machine_id)
        {
            char machine_id_str[16];
            sprintf(machine_id_str, "%u", *(long *)data0);

            res = true;
            *machine_id_2 = machine_id_str;
        }
        else
        {
            vary *_data0 = (vary *)data0;
            _data0->vary_string[_data0->vary_length] = '\0';

            res = true;
            *machine_id_2 = _data0->vary_string;
        }

        retcode = isc_dsql_fetch(status, &st_h, 1, sqlda);
        break;
    }
    if (retcode != 100)
    {
        res = false;
        machine_id_2->clear();

        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_fetch error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        break;
    }

    isc_dsql_free_statement(status, &st_h, DSQL_close);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_free_statement error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        st_h = 0;
        break;
    }
    st_h = 0;

    isc_commit_transaction(status, &tr_h);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_commit_transaction error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        tr_h = 0;
        break;
    }
    tr_h = 0;

    break;} //while (true) {

    if (tr_h)
    {
        isc_rollback_transaction(status, &tr_h);
        if (status[0] == 1 && status[1])
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s isc_rollback_transaction error:\n"
                "%s",
                system_time_str().c_str(),
                firebird_error(status).c_str()
            );
            #endif
        }
    }

    if (st_h)
    {
        isc_dsql_free_statement(status, &st_h, DSQL_close);
        if (status[0] == 1 && status[1])
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s isc_dsql_free_statement error:\n"
                "%s",
                system_time_str().c_str(),
                firebird_error(status).c_str()
            );
            #endif
        }
    }

    free(sqlda);

    delete [] data0;
////////////////////////////////////////////////////////////////////////////////
    #else
////////////////////////////////////////////////////////////////////////////////
    if (mysql_query(&mysql, (std::string() +
        "SELECT `" + (get_virtual_machine_id? "id" : "imei") + "` FROM `machines` "
        "WHERE `" + (get_virtual_machine_id? "imei" : "id") + "` = " +
        (get_virtual_machine_id? "'" : "") + machine_id_1 + (get_virtual_machine_id? "'" : "")).c_str()))
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s mysql_query error: %s\n",
            system_time_str().c_str(),
            mysql_error(&mysql)
        );
        #endif
    }
    else
    {
        MYSQL_RES *mysql_res = mysql_store_result(&mysql);
        if (!mysql_res)
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s mysql_store_result error: %s\n",
                system_time_str().c_str(),
                mysql_error(&mysql)
            );
            #endif
        }
        else
        {
            if (mysql_num_rows(mysql_res) == 1)
            {
                MYSQL_ROW row = mysql_fetch_row(mysql_res);

                res = true;
                *machine_id_2 = row[0];
            }
        }
        mysql_free_result(mysql_res);
    }
////////////////////////////////////////////////////////////////////////////////
    #endif

    return res;
}
//---------------------------------------------------------------------------
static int sizeof_in = 10240;
static int sizeof_out = 1048576;

static int proc_cmd_interval = 2000; //
static int main_interval = 100; //

static bool gc_enabled = false; //garbage collector
static bool q_log_enabled = false; //query log
//---------------------------------------------------------------------------
static std::string app_name;
//---------------------------------------------------------------------------
class simple_server_node;
static void proc(simple_server_node *node, std::vector<std::string> * const data);
static void parse(simple_server_node *node, char *buffer, int * const sizeof_data_cur);
static void proc_cmd(void);
//---------------------------------------------------------------------------
class simple_server_node:
    public se_server_node
{
public:
    int sizeof_in;
    char *in;
    int sizeof_out;
    char *out;

    int parsed;

    std::string machine_id;

    int _sizeof_package;
    int _sizeof_data_cur;

    simple_server_node(void):
        sizeof_in(::sizeof_in),
        in(new char[sizeof_in]),
        sizeof_out(::sizeof_out),
        out(new char[sizeof_out]),
        parsed(0)
    {
        sizeof_buffer = sizeof_in;
        buffer = in;
        sizeof_data = sizeof_buffer;
    }
    virtual ~simple_server_node(void)
    {
        disconnect();

        delete [] in;
        delete [] out;
    }

    virtual void before_job_proc(void)
    {
        if (job_active() && get_job() == _recv && sizeof_data_cur != parsed)
        {
            parse(this, buffer, &sizeof_data_cur);
            parsed = sizeof_data_cur;
        }
    }
    virtual void result_proc(void)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s node (ip = %s, port = %hu) job %s finished with result %s\n",
            system_time_str().c_str(),
            get_ip().c_str(),
            get_port(),
            get_job_str(get_job()),
            get_result_str(get_result())
        );
        #endif

        if (get_result() == aborted) return;

        if (get_result() != success)
        {
            if (get_result() != connection_closed) disconnect();
            return;
        }

        if (get_job() == _send)
        {
            sizeof_buffer = sizeof_in;
            buffer = in;
            sizeof_data = sizeof_buffer;
        }
        else
        {
            parse(this, buffer, &sizeof_data_cur);
            parsed = sizeof_data_cur;

            _sizeof_package = sizeof_package;
            _sizeof_data_cur = sizeof_data_cur;
        }

        if (activate_job(_recv))
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s node (ip = %s, port = %hu) job %s activated\n",
                system_time_str().c_str(),
                get_ip().c_str(),
                get_port(),
                get_job_str(get_job())
            );
            #endif

            sizeof_package = _sizeof_package;
            sizeof_data_cur = _sizeof_data_cur;
        }
        else
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s node (ip = %s, port = %hu) can not activate job %s\n",
                system_time_str().c_str(),
                get_ip().c_str(),
                get_port(),
                get_job_str(get_job())
            );
            #endif
        }
    }
    virtual void after_connect(void)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s node (ip = %s, port = %hu) connected\n",
            system_time_str().c_str(),
            get_ip().c_str(),
            get_port()
        );
        #endif

        if (activate_job(_recv, true))
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s node (ip = %s, port = %hu) job %s activated\n",
                system_time_str().c_str(),
                get_ip().c_str(),
                get_port(),
                get_job_str(get_job())
            );
            #endif
        }
        else
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s node (ip = %s, port = %hu) can not activate job %s\n",
                system_time_str().c_str(),
                get_ip().c_str(),
                get_port(),
                get_job_str(get_job())
            );
            #endif
        }
    }
    virtual void after_disconnect(void)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s node (ip = %s, port = %hu) disconnected\n",
            system_time_str().c_str(),
            get_ip().c_str(),
            get_port()
        );
        #endif
    }

    virtual void after_package(char *package, int sizeof_package)
    {
        #ifdef DEBUG
        printf_ex
        (
            "//---------------------------------------------------------------------------\n"
            "%s node (ip = %s, port = %hu) package (sizeof_package=%i)\n"
            "//---------------------------------------------------------------------------\n"
            "%s\n"
            "//---------------------------------------------------------------------------\n",
            system_time_str().c_str(),
            get_ip().c_str(),
            get_port(),
            sizeof_package,
            std::string(package, sizeof_package).c_str()
        );
        #endif
    }
};

class simple_server:
    public se_server
{
public:
    virtual void after_error(const errors error)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s server function call finished with error %s\n",
            system_time_str().c_str(),
            get_error_str(error)
        );
        #endif
    }

    virtual simple_server_node *new_node(void)
    {
        return new simple_server_node();
    }
};
//---------------------------------------------------------------------------
static simple_server server;
//---------------------------------------------------------------------------
static void proc(simple_server_node *node, std::vector<std::string> * const data)
{
    #ifdef DEBUG
    printf_ex
    (
        "%s node (ip = %s, port = %hu) proc beg\n",
        system_time_str().c_str(),
        node->get_ip().c_str(),
        node->get_port()
    );
    #endif

    std::map<std::string, std::string>::iterator i = params.find((*data)[0]);
    if (i == params.end())
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s params.find error (arg=%s)\n",
            system_time_str().c_str(),
            (*data)[0].c_str()
        );
        #endif
        return;
    }
    std::string table_name = i->second;

    std::string names;
    std::string values;

    for (int idx = 1; idx < data->size(); idx++)
    {
        char buffer[16];
        sprintf(buffer, "%i", idx - 1);
        std::map<std::string, std::string>::iterator i = params.find((*data)[0] + "_data_" + buffer);
        if (i == params.end()) break;

        if (i->second.size() && (*data)[idx].size())
        {
            if (i->second[0] == '`')
            {
                if (i->second == "`machine_id`")
                {
                    node->machine_id = (*data)[idx];

                    std::string virtual_machine_id;
                    if (!get_machine_id((*data)[idx], &virtual_machine_id, true))
                    {
                        #ifdef DEBUG
                        printf_ex
                        (
                            "%s no machine_id for imei=%s\n",
                            system_time_str().c_str(),
                            (*data)[idx].c_str()
                        );
                        #endif
                        return;
                    }
                    (*data)[idx] = virtual_machine_id;
                }

                #ifdef FIREBIRD

                names.append(i->second.c_str() + 1, i->second.size() - 2);
                names.append(",");
                values += "'" + (*data)[idx] + "',";

                #else

                names += i->second + ",";
                values += "'" + (*data)[idx] + "',";

                #endif
            }
            else
            {
                if (i->second == "machine_id")
                {
                    node->machine_id = (*data)[idx];

                    std::string virtual_machine_id;
                    if (!get_machine_id((*data)[idx], &virtual_machine_id, true))
                    {
                        #ifdef DEBUG
                        printf_ex
                        (
                            "%s no machine_id for imei=%s\n",
                            system_time_str().c_str(),
                            (*data)[idx].c_str()
                        );
                        #endif
                        return;
                    }
                    (*data)[idx] = virtual_machine_id;
                }

                #ifdef FIREBIRD

                names += i->second + ",";
                values += (*data)[idx] + ",";

                #else

                names += "`" + i->second + "`,";
                values += (*data)[idx] + ",";

                #endif
            }
        }
    }

    i = params.find((*data)[0] + "_ext_data_count");
    if (i != params.end() && i->second.size())
    {
        int ext_data_count = atoi(i->second.c_str());
        for (int idx = 0; idx < ext_data_count; idx++)
        {
            char buffer[16];
            sprintf(buffer, "%i", idx);
            std::map<std::string, std::string>::iterator i = params.find((*data)[0] + "_ext_data_" + buffer);
            std::map<std::string, std::string>::iterator i_value = params.find((*data)[0] + "_ext_data_" + buffer + "_value");
            if (i == params.end() || i_value == params.end()) continue;

            if (i->second.size())
            {
                if (i->second[0] == '`')
                {
                    #ifdef FIREBIRD

                    names.append(i->second.c_str() + 1, i->second.size() - 2);
                    names.append(",");
                    values += "'" + i_value->second + "',";

                    #else

                    names += i->second + ",";
                    values += "'" + i_value->second + "',";

                    #endif
                }
                else
                {
                    #ifdef FIREBIRD

                    names += i->second + ",";
                    values += i_value->second + ",";

                    #else

                    names += "`" + i->second + "`,";
                    values += i_value->second + ",";

                    #endif
                }
            }
        }
    }

    if (!names.size())
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s names.size=0\n",
            system_time_str().c_str()
        );
        #endif
        return;
    }

    names.resize(names.size() - 1);
    values.resize(values.size() - 1);

    #ifdef FIREBIRD

    std::string query = "INSERT INTO " + table_name + " (" + names + ") VALUES (" + values + ")";

    #else

    std::string query = "INSERT INTO `" + table_name + "` (" + names + ") VALUES (" + values + ")";

    #endif

    #ifdef DEBUG
    if (q_log_enabled) printf_ex("%s\n", query.c_str());
    #endif

    int _good_queries_count = good_queries_count;

    #ifdef FIREBIRD

    if (firebird_exec((char *)query.c_str())) good_queries_count++;

    #else

    if (mysql_query(&mysql, query.c_str()))
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s mysql_query error: %s\n",
            system_time_str().c_str(),
            mysql_error(&mysql)
        );
        #endif
    }
    else
    {
        good_queries_count++;
    }

    #endif

    if (good_queries_count == _good_queries_count) bad_queries_count++;
    else
    {
        std::map<std::string, std::string>::iterator i = params.find((*data)[0] + "_exec");
        if (i != params.end()) system(i->second.c_str());
    }

    #ifdef DEBUG
    printf_ex
    (
        "%s node (ip = %s, port = %hu) proc end\n",
        system_time_str().c_str(),
        node->get_ip().c_str(),
        node->get_port()
    );
    #endif
}
//---------------------------------------------------------------------------
/*

 
$DVSN,359772038050465,280113,082444,S,A,25,31.86,12.08,8.23,0,0,0,1,8.40,8.38,8.43,*A\r


$ - btx_marker
DVSN - table_id
, - data_separator
* - cs_marker
A - cs
\r - etx_marker

*/
//---------------------------------------------------------------------------
static void parse(simple_server_node *node, char *buffer, int * const sizeof_data_cur)
{
    #ifdef DEBUG
    printf_ex
    (
        "%s node (ip = %s, port = %hu) parse beg (sizeof_data_cur=%i)\n",
        system_time_str().c_str(),
        node->get_ip().c_str(),
        node->get_port(),
        *sizeof_data_cur
    );
    #endif

    char *beg = buffer;
    char *end = buffer + *sizeof_data_cur;

    while (true)
    {
        char *btx = find_str(beg, end, (char *)btx_marker.c_str());
        if (!btx)
        {
            *sizeof_data_cur = 0;
            break;
        }
        btx += btx_marker.size();

        char *etx = find_str(btx, end, (char *)etx_marker.c_str());
        if (!etx)
        {
            *sizeof_data_cur -= beg - buffer;
            memmove(buffer, beg, *sizeof_data_cur);
            break;
        }

        beg = etx + etx_marker.size();

        etx -= cs_marker.size() + 2;

        bool cs_extracted = false;

        unsigned short cs;

        if (!memcmp(etx, cs_marker.c_str(), cs_marker.size()))
        {
            etx += cs_marker.size();

            unsigned short lo;
            unsigned short hi;

            if
            (
                hex1_to_ui2(*(unsigned char *)etx, &hi) &&
                hex1_to_ui2(*(unsigned char *)(etx + 1), &lo)
            )
            {
                cs_extracted = true;
                cs = (hi << 4) | lo;

                etx -= cs_marker.size();
            }
        }
        else
        {
            etx++;

            if (!memcmp(etx, cs_marker.c_str(), cs_marker.size()))
            {
                etx += cs_marker.size();

                unsigned short lo;

                if (hex1_to_ui2(*etx, &lo))
                {
                    cs_extracted = true;
                    cs = lo;

                    etx -= cs_marker.size();
                }
            }
        }

        if (cs_extracted)
        {

        #ifdef CS_ENABLED

            if (cs == get_cs((unsigned char *)btx, (unsigned char *)etx))
            {

        #else

            if (true)
            {

        #endif

                std::vector<std::string> data;

                while (true)
                {
                    char *p = find_str(btx, etx, (char *)data_separator.c_str());
                    if (!p)
                    {
                        data.push_back(std::string(btx, etx - btx));
                        break;
                    }

                    data.push_back(std::string(btx, p - btx));

                    btx = p + data_separator.size();
                }

                proc(node, &data);
            }
        }
    }

    #ifdef DEBUG
    printf_ex
    (
        "%s node (ip = %s, port = %hu) parse end (sizeof_data_cur=%i)\n",
        system_time_str().c_str(),
        node->get_ip().c_str(),
        node->get_port(),
        *sizeof_data_cur
    );
    #endif
}
//---------------------------------------------------------------------------
/*

      commands

 
id - INTEGER -  ,      
machine_id - VARCHAR -   -   . 
text - VARCHAR -   -  
status - INTEGER -  ,    0
0 -    
1 -   

        status   1
           machine_id,  machine_id  

*/
//---------------------------------------------------------------------------
static void proc_cmd(void)
{
    #ifdef FIREBIRD
////////////////////////////////////////////////////////////////////////////////
    isc_tr_handle tr_h = 0;

    isc_stmt_handle st_h = 0;

    XSQLDA *sqlda = (XSQLDA *)malloc(XSQLDA_LENGTH(3));
    sqlda->sqln = 3;
    sqlda->version = 1;

    char *data0 = 0, *data1 = 0, *data2 = 0;

    while (true) {

    isc_start_transaction(status, &tr_h, 1, &db_h, 0, 0);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_start_transaction error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        tr_h = 0;
        break;
    }

    isc_dsql_allocate_statement(status, &db_h, &st_h);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_allocate_statement error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        st_h = 0;
        break;
    }

    isc_dsql_prepare(status, &tr_h, &st_h, 0, "SELECT id, machine_id, text FROM commands WHERE status = 0", 1, sqlda);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_prepare error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        break;
    }

    short ind0 = 0, ind1 = 0, ind2 = 0;

    //   :
    //2 -  vary_length
    //sqllen - .  vary_string
    //1 -  

    sqlda->sqlvar[0].sqldata = data0 = new char[sizeof(long)];
    sqlda->sqlvar[0].sqltype = SQL_LONG + 1;
    sqlda->sqlvar[0].sqlind  = &ind0;

    sqlda->sqlvar[1].sqldata = data1 = new char[sizeof(long)];
    sqlda->sqlvar[1].sqltype = SQL_LONG + 1;
    sqlda->sqlvar[1].sqlind  = &ind1;

    sqlda->sqlvar[2].sqldata = data2 = new char[2 + sqlda->sqlvar[2].sqllen + 1];
    sqlda->sqlvar[2].sqltype = SQL_VARYING + 1;
    sqlda->sqlvar[2].sqlind  = &ind2;
    vary *_data2 = (vary *)data2;

    isc_dsql_execute(status, &tr_h, &st_h, 1, 0);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_execute error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        break;
    }

    ISC_STATUS retcode;
    while (true)
    {
        retcode = isc_dsql_fetch(status, &st_h, 1, sqlda);
        if (retcode) break;

        char id_str[16];
        sprintf(id_str, "%u", *(long *)data0);

        char virtual_machine_id_str[16];
        sprintf(virtual_machine_id_str, "%u", *(long *)data1);

        _data2->vary_string[_data2->vary_length] = '\0';

        std::string machine_id;
        if (!get_machine_id(virtual_machine_id_str, &machine_id, false))
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s no imei for machine_id=%s\n",
                system_time_str().c_str(),
                virtual_machine_id_str
            );
            #endif
            continue;
        }

        for
        (
            std::list<se_server_node *>::iterator p = server.nodes.begin();
            p != server.nodes.end();
            p++
        )
        {
            simple_server_node *_p = dynamic_cast<simple_server_node *>(*p);
            if (_p->machine_id == machine_id)
            {
                if (_p->get_job() == se_basic_node::_recv)
                {
                    std::string query = "UPDATE commands SET status = 1 WHERE id = ";
                    query += id_str;

                    if (firebird_exec((char *)query.c_str()))
                    {
                        _p->_sizeof_package = _p->sizeof_package;
                        _p->_sizeof_data_cur = _p->sizeof_data_cur;

                        _p->sizeof_buffer = _p->sizeof_out;
                        _p->buffer = _p->out;
                        _p->sizeof_data = sprintf(_p->out, "%s", _data2->vary_string);

                        _p->abort_job();

                        if (_p->activate_job(se_basic_node::_send))
                        {
                            #ifdef DEBUG
                            printf_ex
                            (
                                "%s node (ip = %s, port = %hu) job %s activated\n",
                                system_time_str().c_str(),
                                _p->get_ip().c_str(),
                                _p->get_port(),
                                _p->get_job_str(_p->get_job())
                            );
                            #endif
                        }
                        else
                        {
                            #ifdef DEBUG
                            printf_ex
                            (
                                "%s node (ip = %s, port = %hu) can not activate job %s\n",
                                system_time_str().c_str(),
                                _p->get_ip().c_str(),
                                _p->get_port(),
                                _p->get_job_str(_p->get_job())
                            );
                            #endif
                        }
                    }
                }
                break;
            }
        }
    }
    if (retcode != 100)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_fetch error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        break;
    }

    isc_dsql_free_statement(status, &st_h, DSQL_close);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_dsql_free_statement error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        st_h = 0;
        break;
    }
    st_h = 0;

    isc_commit_transaction(status, &tr_h);
    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_commit_transaction error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
        tr_h = 0;
        break;
    }
    tr_h = 0;

    break;} //while (true) {

    if (tr_h)
    {
        isc_rollback_transaction(status, &tr_h);
        if (status[0] == 1 && status[1])
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s isc_rollback_transaction error:\n"
                "%s",
                system_time_str().c_str(),
                firebird_error(status).c_str()
            );
            #endif
        }
    }

    if (st_h)
    {
        isc_dsql_free_statement(status, &st_h, DSQL_close);
        if (status[0] == 1 && status[1])
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s isc_dsql_free_statement error:\n"
                "%s",
                system_time_str().c_str(),
                firebird_error(status).c_str()
            );
            #endif
        }
    }

    free(sqlda);

    delete [] data0;
    delete [] data1;
    delete [] data2;
////////////////////////////////////////////////////////////////////////////////
    #else
////////////////////////////////////////////////////////////////////////////////
    if (mysql_query(&mysql, "SELECT `id`, `machine_id`, `text` FROM `commands` WHERE `status` = 0"))
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s mysql_query error: %s\n",
            system_time_str().c_str(),
            mysql_error(&mysql)
        );
        #endif
    }
    else
    {
        MYSQL_RES *res = mysql_store_result(&mysql);
        if (!res)
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s mysql_store_result error: %s\n",
                system_time_str().c_str(),
                mysql_error(&mysql)
            );
            #endif
        }
        else
        {
            for (int i = 0; i < mysql_num_rows(res); i++)
            {
                MYSQL_ROW row = mysql_fetch_row(res);

                std::string machine_id;
                if (!get_machine_id(row[1], &machine_id, false))
                {
                    #ifdef DEBUG
                    printf_ex
                    (
                        "%s no imei for machine_id=%s\n",
                        system_time_str().c_str(),
                        row[1]
                    );
                    #endif
                    continue;
                }

                for
                (
                    std::list<se_server_node *>::iterator p = server.nodes.begin();
                    p != server.nodes.end();
                    p++
                )
                {
                    simple_server_node *_p = dynamic_cast<simple_server_node *>(*p);
                    if (_p->machine_id == machine_id)
                    {
                        if (_p->get_job() == se_basic_node::_recv)
                        {
                            std::string query = "UPDATE `commands` SET `status` = 1 WHERE `id` = ";
                            query += row[0];

                            if (mysql_query(&mysql, query.c_str()))
                            {
                                #ifdef DEBUG
                                printf_ex
                                (
                                    "%s mysql_query error: %s\n",
                                    system_time_str().c_str(),
                                    mysql_error(&mysql)
                                );
                                #endif
                            }
                            else
                            {
                                _p->_sizeof_package = _p->sizeof_package;
                                _p->_sizeof_data_cur = _p->sizeof_data_cur;

                                _p->sizeof_buffer = _p->sizeof_out;
                                _p->buffer = _p->out;
                                _p->sizeof_data = sprintf(_p->out, "%s", row[2]);

                                _p->abort_job();

                                if (_p->activate_job(se_basic_node::_send))
                                {
                                    #ifdef DEBUG
                                    printf_ex
                                    (
                                        "%s node (ip = %s, port = %hu) job %s activated\n",
                                        system_time_str().c_str(),
                                        _p->get_ip().c_str(),
                                        _p->get_port(),
                                        _p->get_job_str(_p->get_job())
                                    );
                                    #endif
                                }
                                else
                                {
                                    #ifdef DEBUG
                                    printf_ex
                                    (
                                        "%s node (ip = %s, port = %hu) can not activate job %s\n",
                                        system_time_str().c_str(),
                                        _p->get_ip().c_str(),
                                        _p->get_port(),
                                        _p->get_job_str(_p->get_job())
                                    );
                                    #endif
                                }
                            }
                        }
                        break;
                    }
                }
            }
        }
        mysql_free_result(res);
    }
////////////////////////////////////////////////////////////////////////////////
    #endif
}
//---------------------------------------------------------------------------
static void gc(void)
{
    std::map<std::string, se_server_node *> machine_ids;

    for
    (
        std::list<se_server_node *>::iterator i = server.nodes.begin();
        i != server.nodes.end();
        i++
    )
    {
        std::string machine_id = ((simple_server_node *)(*i))->machine_id;
        if (machine_id.size())
        {
            std::map<std::string, se_server_node *>::iterator p = machine_ids.find(machine_id);
            if (p == machine_ids.end())
            {
                machine_ids.insert(std::map<std::string, se_server_node *>::value_type(machine_id, *i));
            }
            else
            {
                p->second->disconnect();
                p->second = *i;
            }
        }
    }
}
//---------------------------------------------------------------------------
class simple_thread_obj:
    public thread_obj
{
    virtual void main(void)
    {
        /*#ifdef IS_LINUX
        struct timeval tv;
        struct timezone tz;
        gettimeofday(&tv, &tz);
        printf_ex
        (
            "thread main cycle time: %is %ims\n",
            (int)tv.tv_sec, (int)tv.tv_usec / 1000
        );
        #endif*/

        static std::time_t proc_cmd_time = std::time(0);

        server.lock();
        server.job_proc();
        if (gc_enabled) gc();
        if (std::difftime(std::time(0), proc_cmd_time) * 1000 >= proc_cmd_interval)
        {
            proc_cmd_time = std::time(0);
            proc_cmd();
        }
        server.unlock();

        Sleep(main_interval);
    }
};
//---------------------------------------------------------------------------
static simple_thread_obj thread;
//---------------------------------------------------------------------------
static void go_to_next_line(void)
{
    while (true)
    {
        char c;
        int res = scanf("%c", &c);
        if (!res || res == EOF || c == '\n') break;
    }
}
//---------------------------------------------------------------------------
static void before_exit(void)
{
    server.deactivate();

    #ifndef IS_LINUX

    WSACleanup();

    #endif

    #ifdef DEBUG
    printf_ex
    (
        "%s sockets uninitialized\n",
        system_time_str().c_str()
    );
    #endif

    #ifdef FIREBIRD

    isc_detach_database(status, &db_h);

    #else

    mysql_close(&mysql);

    #endif

    #ifdef DEBUG
    printf_ex
    (
        "%s mysql closed\n",
        system_time_str().c_str()
    );
    #endif

    printf_ex_uninit();
}
//---------------------------------------------------------------------------
static void sig_handler(int sig)
{
    thread.destroy();
}
//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
    #ifdef DEBUG
    bool trace_mode = argc == 2 && (!strcmp(argv[1], "-t") || !strcmp(argv[1], "--trace"));
    #endif

    bool interactive_mode = argc == 2 && (!strcmp(argv[1], "-i") || !strcmp(argv[1], "--interactive"));

    #ifdef IS_LINUX
    bool daemon_mode = argc == 2 && (!strcmp(argv[1], "-d") || !strcmp(argv[1], "--daemon"));
    #endif

    bool help_mode = argc == 2 && (!strcmp(argv[1], "-h") || !strcmp(argv[1], "--help"));

    app_name = get_app_name();

    #ifdef DEBUG
    if (trace_mode) //trace mode
    {}
    else
    #endif

    if (interactive_mode) //interactive mode
    {}
    else

    #ifdef IS_LINUX
    if (daemon_mode) //daemon mode
    {
        if (getppid() != 1)
        {
            pid_t pid = fork();
            if (pid < 0)
            {
                return -1;
            }
            if (pid > 0)
            {
                return 0;
            }

            umask(0);

            pid_t sid = setsid();
            if (sid < 0)
            {
                return -2;
            }

            if (chdir("/") < 0)
            {
                return -3;
            }

            for (int i = getdtablesize() - 1; i >= 0; i--)
                close(i);
            open("/dev/null", O_RDONLY); //stdin
            open("/dev/null", O_WRONLY); //stdout
            dup(1); //stderr
        }
    }
    else
    #endif

    {
        printf
        (
            "Syntax:\n"
            "app [option]\n"

            #ifdef DEBUG
            "-t --trace\n"
            "\ttrace mode\n"
            #endif

            "-i --interactive\n"
            "\tinteractive mode\n"

            #ifdef IS_LINUX
            "-d --daemon\n"
            "\tdaemon mode\n"
            #endif

            "-h --help\n"
            "\thelp mode\n"
        );
        
        return help_mode? 0 : -1;
    }

    #ifdef DEBUG
    printf_ex_init(true, trace_mode, change_ext(app_name, ".log").c_str(), "w");
    printf_ex
    (
        "//---------------------------------------------------------------------------\n"
        "%s logging started\n",
        system_time_str().c_str()
    );
    #endif

    std::ifstream is(change_ext(app_name, ".cfg").c_str());
    std::string params_str;
    if (!is)
    {
        printf_ex("////can not read cfg-file\n");
    }
    else
    {
        while (is.peek() != EOF)
        {
            std::string line;
            std::getline(is, line);
            if (line.length() && line.at(line.length() - 1) == '\r') line.resize(line.length() - 1);
            if (line.length() && (line.length() < 2 || line.at(0) != '/' || line.at(1) != '/'))
            {
                params_str.append(line);
                params_str.append("\r\n");
            }
        }
        params_str.append("\r\n");
        is.close();
    }

    std::string ip = "0";
    unsigned short port = 80;

    char *params_s = (char *)params_str.c_str();
    std::map<std::string, std::string>::iterator i;
    if (!parse_params(params_s, params_s + params_str.length(), &params))
    {
        printf_ex("////can not parse params\n");
    }
    else
    {
        i = params.find("ip");
        if (i != params.end()) ip = i->second;

        i = params.find("port");
        if (i != params.end()) port = atoi(i->second.c_str());

        i = params.find("sizeof_in");
        if (i != params.end()) sizeof_in = atoi(i->second.c_str());

        i = params.find("sizeof_out");
        if (i != params.end()) sizeof_out = atoi(i->second.c_str());

        i = params.find("data_separator");
        if (i != params.end()) data_separator = i->second;

        i = params.find("db_address");
        if (i != params.end()) db_address = i->second;

        i = params.find("db_port");
        if (i != params.end()) db_port = atoi(i->second.c_str());

        i = params.find("db_login");
        if (i != params.end()) db_login = i->second;

        i = params.find("db_password");
        if (i != params.end()) db_password = i->second;

        i = params.find("db_name");
        if (i != params.end()) db_name = i->second;

        i = params.find("proc_cmd_interval");
        if (i != params.end()) proc_cmd_interval = atoi(i->second.c_str());

        i = params.find("main_interval");
        if (i != params.end()) main_interval = atoi(i->second.c_str());

        i = params.find("gc_enabled");
        if (i != params.end()) gc_enabled = i->second == "true";

        i = params.find("q_log_enabled");
        if (i != params.end()) q_log_enabled = i->second == "true";
    }

    #ifdef FIREBIRD

    char *dpb = 0;
    short dpb_length = 0;
    isc_expand_dpb(&dpb, &dpb_length, isc_dpb_user_name, db_login.c_str(), isc_dpb_password, db_password.c_str(), 0);

    char db_port_str[16];
    sprintf(db_port_str, "%hu", db_port);

    std::string full_db_name = db_address + "/" + db_port_str + ":" + db_name;
    isc_attach_database(status, 0, (char *)full_db_name.c_str(), &db_h, dpb_length, dpb);

    if (status[0] == 1 && status[1])
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s isc_attach_database error:\n"
            "%s",
            system_time_str().c_str(),
            firebird_error(status).c_str()
        );
        #endif
    }
    else
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s database attached\n",
            system_time_str().c_str()
        );
        #endif
    }

    #else

    if (!mysql_init(&mysql))
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s mysql_init error: %s\n",
            system_time_str().c_str(),
            mysql_error(&mysql)
        );
        #endif
    }
    else
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s mysql initialized\n",
            system_time_str().c_str()
        );
        #endif
    }

    if (!mysql_real_connect(&mysql, db_address.c_str(), db_login.c_str(), db_password.c_str(), db_name.c_str(), db_port, 0, 0))
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s mysql_real_connect error: %s\n",
            system_time_str().c_str(),
            mysql_error(&mysql)
        );
        #endif
    }
    else
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s mysql connected\n",
            system_time_str().c_str()
        );
        #endif
    }

    #endif

    #ifndef IS_LINUX

    WSADATA wsa_data;
    if (WSAStartup(0x101, &wsa_data) || wsa_data.wVersion != 0x101)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s can not initialize sockets\n",
            system_time_str().c_str()
        );
        #endif
    }
    else
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s sockets initialized\n",
            system_time_str().c_str()
        );
        #endif
    }

    #endif

    server.prepare
    (
        AF_INET,
        SOCK_STREAM,
        IPPROTO_TCP,
        ip.c_str(),
        port
    );
    if (!server.activate(true))
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s can not activate server\n",
            system_time_str().c_str()
        );
        #endif
    }
    else
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s server activated\n",
            system_time_str().c_str()
        );
        #endif
    }

    thread.create();

    signal(SIGINT, sig_handler);
    signal(SIGTERM, sig_handler);

    while (true)
    {
        /*#ifdef IS_LINUX
        struct timeval tv;
        struct timezone tz;
        gettimeofday(&tv, &tz);
        printf_ex
        (
            "app main cycle time: %is %ims\n",
            (int)tv.tv_sec, (int)tv.tv_usec / 1000
        );
        #endif*/

        if (thread.terminated()) break;

        if (!interactive_mode)
        {
            Sleep(1000);
            continue;
        }

        printf(">");
        char cmd[256];
        cmd[0] = '\0';
        scanf("%255s", cmd);
        if (!strcmp(cmd, "exit"))
        {
            thread.destroy();
            break;
        }
        if (!strcmp(cmd, "help"))
        {
            printf
            (
                "Commands:\n"
                "\texit\n"
                "\thelp\n"
                "\tinfo\n"
                "\tdisconnect ip (or *) port (or *)\n"
            );
        }
        else if (!strcmp(cmd, "info"))
        {
            server.lock();
            printf("//---------------------------------------------------------------------------\n");
            printf
            (
                "active: %s\n"
                "nodes count: %i\n"
                "good queries count: %i\n"
                "bad queries count: %i\n",
                server.active()? "true" : "false",
                server.nodes.size(),
                good_queries_count,
                bad_queries_count
            );
            if (server.nodes.size())
            {
                printf
                (
                    "\n"
                    "%-11s " "%-6s " "%-25s " "%-16s " "%s\n",
                    "job", "active", "job_activation_time", "ip", "port"
                );
                for
                (
                    std::list<se_server_node *>::iterator p = server.nodes.begin();
                    p != server.nodes.end();
                    p++
                )
                    printf
                    (
                        "%-11s " "%-6s " "%-25s " "%-16s " "%hu\n",
                        (*p)->get_job_str((*p)->get_job()),
                        (*p)->job_active()? "true" : "false",
                        system_time_str((*p)->job_activation_time()).c_str(),
                        (*p)->get_ip().c_str(),
                        (*p)->get_port()
                    );
            }
            printf("//---------------------------------------------------------------------------\n");
            server.unlock();
        }
        else if (!strcmp(cmd, "disconnect"))
        {
            ip.clear();
            port = 0;
            
            cmd[0] = '\0';
            scanf("%255s", cmd);
            if (cmd[0] != '*') ip = cmd;

            cmd[0] = '\0';
            scanf("%255s", cmd);
            if (cmd[0] != '*') port = atoi(cmd);

            int c = 0;

            server.lock();
            for
            (
                std::list<se_server_node *>::iterator p = server.nodes.begin();
                p != server.nodes.end();
                p++
            )
                if
                (
                    (!ip.size() || (*p)->get_ip() == ip) &&
                    (!port || (*p)->get_port() == port)
                )
                {
                    (*p)->disconnect();
                    c++;
                }
            server.unlock();

            printf("disconnected: %i\n", c);
        }
        else
        {
            printf("unknown command\n");
        }
        go_to_next_line();
    }

    before_exit();

    return 0;
}
//---------------------------------------------------------------------------

