//------------------------------------------------------------------------------
#ifndef socket_engine_h
#define socket_engine_h
//------------------------------------------------------------------------------
//Copyright (C) 2013 by Gusev Alexander
//e-mail: alphacity@rambler.ru
//------------------------------------------------------------------------------
#include <string>
#include <list>
#include <ctime>
//------------------------------------------------------------------------------
#include "platform.h"
#include "mt_obj.h"
//------------------------------------------------------------------------------
// -;      -
//------------------------------------------------------------------------------
class se_basic_node;
//------------------------------------------------------------------------------
bool prepare_server_node(se_basic_node * const obj, SOCKET s, SOCKADDR_IN sa); //   ;  _connected   true;     -  false
bool prepare_client_node //        ;     -  false
(
    se_basic_node * const obj,
    const int af,
    const int type,
    const int protocol,
    const char *ip,
    const unsigned short port
);
//------------------------------------------------------------------------------
class se_basic_node
{
public:
    enum jobs
    {
        no_job = 0,
        _connect,
        _shutdown,
        _disconnect, //      _disconnect     _shutdown
        _recv,
        _send,
        unknown_job,
        countof_jobs
    };

    enum results
    {
        no_result = 0, 
        busy, // 
        already_connected,
        socket_failed,
        can_not_enable_nonblocking_mode,
        select_failed,
        connection_attempt_failed,
        already_disconnected,
        shutdown_failed,
        recv_failed,
        connection_closed, //  
        send_failed,
        aborted_with_timeout,
        aborted,
        success,
        success_with_overflow, //   (     / );  - sizeof_data > sizeof_buffer
        unknown_result,
        countof_results
    };

    unsigned long job_timeout; //   , 

    bool package_mode; //   ( _recv        )

    int sizeof_package; //  / 
    int sizeof_data_cur; //    /
    int sizeof_data; //    /
    int sizeof_buffer; // 
    char *buffer; //   /   ;    

    se_basic_node(void);
    virtual ~se_basic_node(void); //   ;  - -    ,  - disconnect        - disconnect    

    bool activate_job(const jobs job, const bool ready_for_job_proc = false); //        -  false;  ready_for_job_proc      - result_proc, after_connect  after_disconnect;  ready_for_job_proc      
    void abort_job(void);
    bool job_active(void);
    std::time_t job_activation_time(void); //  
    jobs get_job(void);
    static const char *get_job_str(const jobs job);
    results get_result(void);
    static const char *get_result_str(const results result);
    bool ready_for_job_proc(void); //  - before_job_proc  job_proc     - result_proc (         )

    void job_proc(void); // 

    virtual void before_job_proc(void) {} //     // 
    virtual void result_proc(void) {} //  (   )
    virtual void after_connect(void) {}
    virtual void after_disconnect(void) {}

    void disconnect(void); //  (  )
    bool connected(void);

    std::string get_ip(void);
    unsigned short get_port(void);

    virtual void after_package(char *package, int sizeof_package) {} //  

private:
    //     (      se_server_node)
    int af;
    int type;
    int protocol;

    SOCKET s;
    SOCKADDR_IN sa;

    bool _job_active;
    std::time_t _job_activation_time;
    jobs job;
    static const char *job_str[countof_jobs];
    results result;
    static const char *result_str[countof_results];
    bool _ready_for_job_proc;

    bool _connected;

    se_basic_node(const se_basic_node &) {}
    const se_basic_node &operator=(const se_basic_node &) {}

    friend
    bool prepare_server_node(se_basic_node * const obj, SOCKET s, SOCKADDR_IN sa);

    friend
    bool prepare_client_node
    (
        se_basic_node * const obj,
        const int af,
        const int type,
        const int protocol,
        const char *ip,
        const unsigned short port
    );
};
//------------------------------------------------------------------------------
class se_server_node:
    public se_basic_node
{};
//------------------------------------------------------------------------------
class se_client_node:
    public se_basic_node
{
public:
    bool prepare //        ;     -  false
    (
        const int af,
        const int type,
        const int protocol,
        const char *ip,
        const unsigned short port
    );
};
//------------------------------------------------------------------------------
class se_server:
    public mt_obj
{
public:
    enum errors
    {
        no_error = 0,
        socket_failed,
        bind_failed,
        listen_failed,
        can_not_enable_nonblocking_mode,
        accept_failed,
        can_not_enable_nonblocking_mode_on_accepted_socket,
        select_failed,
        unknown_error,
        countof_errors
    };

    std::list<se_server_node *> nodes; //    

    se_server(void);
    virtual ~se_server(void); //  ;  - -    ,  - deactivate        - deactivate    

    bool prepare //        ;     -  false
    (
        const int af,
        const int type,
        const int protocol,
        const char *ip,
        const unsigned short port
    );
    bool activate(const bool need_for_listen); // ;       ;         -  false
    void deactivate(void); // ;         ;   
    bool active(void);

    void job_proc(void); // 

    virtual void after_error(const errors error) {}

    static const char *get_error_str(const errors error);

    std::string get_ip(void); // ip  "0.0.0.0"     
    unsigned short get_port(void);

    virtual se_server_node *new_node(void); //-      

private:
    //    
    int af;
    int type;
    int protocol;

    SOCKET s;
    SOCKADDR_IN sa;

    bool _active;

    static const char *error_str[countof_errors];

    se_server(const se_server &) {}
    const se_server &operator=(const se_server &) {}
};
//------------------------------------------------------------------------------
class se_client:
    public mt_obj
{
public:
    std::list<se_client_node *> nodes; //   ;              

    se_client(void);
    virtual ~se_client(void); //      ,    ;  - -    ,  - disconnect  delete_all_nodes        - disconnect  delete_all_nodes    

    void job_proc(void); // 

    void disconnect(void); //   
    void delete_all_nodes(void); //  ,   

private:
    se_client(const se_client &) {}
    const se_client &operator=(const se_client &) {}
};
//------------------------------------------------------------------------------
#endif //socket_engine_h
