//------------------------------------------------------------------------------
#include "thread_obj.h"
//------------------------------------------------------------------------------
#ifndef IS_LINUX

static DWORD WINAPI thread_obj_main(LPVOID param)

#else

static void *thread_obj_main(void *param)

#endif
{
    thread_obj *p = (thread_obj *)param;

    while (true)
    {
        if (p->terminated()) break;

        p->main();
    }

    #ifndef IS_LINUX

    ExitThread(0);

    #else

    pthread_exit(0);

    #endif
}
//------------------------------------------------------------------------------
bool thread_obj::create(void)
{
    bool res = false;

    lock();
    if (!_terminated)
    {
        unlock();
    }
    else
    {
        _terminated = false;
        unlock();

        #ifndef IS_LINUX

        DWORD thread_id;
        thread_h = CreateThread
        (
            0,
            0,
            thread_obj_main,
            this,
            0,
            &thread_id
        );
        if (thread_h) res = true;
        else _terminated = true;

        #else

        if (!pthread_create(&thread_id, 0, thread_obj_main, this)) res = true;
        else _terminated = true;

        #endif
    }

    return res;
}
//------------------------------------------------------------------------------
void thread_obj::destroy(void)
{
    lock();
    if (_terminated)
    {
        unlock();
    }
    else
    {
        _terminated = true;
        unlock();

        #ifndef IS_LINUX

        WaitForSingleObject(thread_h, INFINITE);
        CloseHandle(thread_h);
        thread_h = 0;

        #else

        pthread_join(thread_id, 0);
        thread_id = 0;

        #endif
    }
}
//------------------------------------------------------------------------------
bool thread_obj::terminated(void)
{
    lock();
    bool res = _terminated;
    unlock();
    return res;
}
