//////////////////////////////////////////////////////////////////////////
/// func lib for gps_board
//////////////////////////////////////////////////////////////////////////
#include <avr\io.h>
#include <avr\interrupt.h>
#include <avr\eeprom.h>
#include <avr\pgmspace.h>
#include <avr\delay.h>
#include "func.h"
#include "..\extern.h"
#include "..\defines.h"

volatile uint16_t millstim = 0; //  dly_ms init on gnss_board.c

//////////////////////////////////////////////////////////////////////////
// DELAY
void dly_timer_decrement(void)
{
	if(millstim != 0)
	{
		millstim--;
	}
}

void dly_ms(uint16_t _d)
{
	millstim = _d;
	while(millstim != 0) Idle();
}

//////////////////////////////////////////////////////////////////////////
// SPI
//////////////////////////////////////////////////////////////////////////
void spi_init(void)
{
	PR.PRPD &= ~0x08; // to get SPID back from sleep
	SPID.CTRL = SPI_ENABLE_bm | SPI_MASTER_bm | SPI_MODE_0_gc | SPI_PRESCALER_DIV64_gc;
}

void spi_wait(void)
{
	while(!(SPID.STATUS & (1<<7)));
}

void spi_send_byte(uint8_t d)
{
	SPID.DATA = d;
	spi_wait();
}

uint8_t spi_read_byte(void)
{
	spi_send_byte(0xFF);
	return SPID.DATA;
}

void spi_send_str(const char *str)
{
	while(*str)
	spi_send_byte(*(str++));
}
//////////////////////////////////////////////////////////////////////////
//TWI functions
//////////////////////////////////////////////////////////////////////////

#define CPU_SPEED 32000000
#define BAUDRATE    400000
#define TWI_BAUD(F_SYS, F_TWI) ((F_SYS / (2 * F_TWI)) - 5)
#define TWI_BAUDSETTING TWI_BAUD(CPU_SPEED, BAUDRATE)

void twi_init(void)
{
	TWIE.MASTER.CTRLA = TWI_MASTER_ENABLE_bm | TWI_MASTER_WIEN_bm | TWI_MASTER_ENABLE_bm;
	TWIE.MASTER.CTRLB = TWI_MASTER_INTLVL_LO_gc | TWI_MASTER_WIF_bm | TWI_MASTER_RIF_bm | TWI_MASTER_SMEN_bm;
	TWIE.MASTER.BAUD =  TWI_BAUDSETTING;
	TWIE.MASTER.STATUS = TWI_MASTER_BUSSTATE_IDLE_gc;
}

void twi_send_byte(uint8_t _addr, uint8_t _byte)
{
	TWIE.MASTER.ADDR = _addr & ~0x01;
	TWIE.MASTER.DATA = _byte;
	while (!(TWIE.MASTER.STATUS&TWI_MASTER_WIF_bm));
}

uint8_t twi_read_byte(uint8_t _addr)
{
	TWIE.MASTER.ADDR = _addr | 0x01;
	while(!(TWIE.MASTER.STATUS&TWI_MASTER_RIF_bm));
	return TWIE.MASTER.DATA;	
}

void twi_send_str(uint8_t _addr, const char *str)
{
	while(*str)
	twi_send_byte(_addr, *(str++));
}
//twi interrupt
ISR(TWIE_TWIM_vect)
{
	rs485_send_byte(twi_read_byte(0x0d));
}

//////////////////////////////////////////////////////////////////////////
// STRING FUNCTIONS
//////////////////////////////////////////////////////////////////////////
/*
char *num_to_string(uint32_t _v, uint8_t _w)
{
	switch (_w) {
		case 0:  sprintf(_num_to_str_buff_i, "%u",		(uint8_t)_v); break;//number
		case 1:  sprintf(_num_to_str_buff_i, "%01u",	(uint8_t)_v); break; //0x
		case 2:  sprintf(_num_to_str_buff_i, "%02u",	(uint16_t)_v); break; //00x
		case 3:  sprintf(_num_to_str_buff_i, "%03u",	(uint16_t)_v); break; //000x
		case 4:  sprintf(_num_to_str_buff_i, "%04u",	(uint32_t)_v); break; //0000x
		case 5:  sprintf(_num_to_str_buff_i, "%05u",	(uint32_t)_v); break; //00000x
		case 6:  sprintf(_num_to_str_buff_i, "%06lu",	(uint32_t)_v); break; //000000x
		case 11: sprintf(_num_to_str_buff_i, "%c",		(uint8_t)_v); //char
	}
	return _num_to_str_buff_i;
}
*/

//getting check summ for string
uint16_t getCheckSum(char *str)
{
	int i;
	int XOR;
	int c;
	// Calculate checksum ignoring any $'s in the string
	for (XOR = 0, i = 0; i < strlen(str); i++) {
		c = (unsigned char)str[i];
		if (c == '*') break;
		if (c != '$') XOR ^= c;
	}
	return XOR;
}

//////////////////////////////////////////////////////////////////////////
//
//   
/*
void queue_add_string(char *_to_rx_buffer, char *_command, char *_count) // ,  ,  
{
	if(strlen(_command) + strlen(_to_rx_buffer) + 1 < QUEUE_RX_BUFF_SIZE) //   
	{
		memcpy(_to_rx_buffer+strlen(_to_rx_buffer), _command, strlen(_command)); //   
		_to_rx_buffer[strlen(_to_rx_buffer)] = QUEUE_SEP; // 
		_count++; //  
	}
}*/

//    ( !      )
void queue_add_byte(char *to_rx_buffer, char *_byte) // ,   
{
	char *_buff_end = to_rx_buffer + QUEUE_RX_BUFF_SIZE - 1;
	while(*to_rx_buffer)
		*(to_rx_buffer++);
	if(to_rx_buffer < _buff_end)
		memcpy(to_rx_buffer, &_byte, 1);
	else
	{
		memset(to_rx_buffer, '\0', QUEUE_RX_BUFF_SIZE); // 
		//rs485_send_byte(_byte);
		//rs485_send_str_P(PSTR("QUBUFF_OVF! RESET\r\n"));	
		//CPU_CCP=CCP_IOREG_gc;
		//RST.CTRL=RST_SWRST_bm;	
	}
}

//     
void queue_read(char *to_rx_buffer, char *_command) // ,  
{
	memset(_command, '\0', QUEUE_COMMAND_BUFF_SIZE); //    
	memccpy(_command, to_rx_buffer, QUEUE_SEP, QUEUE_COMMAND_BUFF_SIZE - 1); //         
	uint16_t _newsize = strlen(to_rx_buffer) - strlen(_command); //     
	memcpy(to_rx_buffer, to_rx_buffer + strlen(_command),  _newsize); //      
	memset(to_rx_buffer +_newsize, '\0', QUEUE_RX_BUFF_SIZE - _newsize); //  eof
	_command[strlen(_command) - 1] = 0x00; //      
}

//   eeprom
char *eeprom_read_str(uint16_t _addr)
{
	char _eeprom_buff[110];
	//  
	memset(_eeprom_buff, '\0', 110);
	//       eeprom   \0
	char _b;
	uint8_t p_str_buff = 0;
	while((_b = eeprom_read_byte(_addr+p_str_buff)) != 0)
	{	
		memcpy(_eeprom_buff+p_str_buff, &_b, 1);
		p_str_buff++;
	}
	//   
	return _eeprom_buff;
}

//////////////////////////////////////////////////////////////////////////
//   

//  
uint8_t ReadCalibrationByte( uint8_t index )
{
	uint8_t result;
	/* Load the NVM Command register to read the calibration row. */
	NVM_CMD = NVM_CMD_READ_CALIB_ROW_gc;
	result = pgm_read_byte(index);
	/* Clean up NVM Command register. */
	NVM_CMD = NVM_CMD_NO_OPERATION_gc;
	return( result );
}

float ReadECUtemp(void)
{
	float degrees_x10;
	float kelvin_per_adc_x10;
	
	kelvin_per_adc_x10 = ((273 + 85)*10) / (float)system.temp_reff; // reference is ADC reading at 85C, scaled by 10 to get units of 0.1C
	
	ADCA.CTRLA = ADC_ENABLE_bm;// enable adc
	
	ADCA.CTRLB = ADC_RESOLUTION_12BIT_gc;                // 12 bit conversion
	ADCA.REFCTRL = ADC_REFSEL_INT1V_gc | ADC_TEMPREF_bm;   // internal 1V bandgap reference + Temp Sensor enable //
	ADCA.PRESCALER = ADC_PRESCALER_DIV256_gc;             // peripheral clk/256 (32MHz/256=125kHz)
	
	ADCA.CH0.CTRL = ADC_CH_INPUTMODE_INTERNAL_gc | ADC_CH_GAIN_1X_gc;  // Internal Temp Sensor
	ADCA.CH0.MUXCTRL = ADC_CH_MUXINT_TEMP_gc;

	for(uint8_t _x=0; _x<10; _x++)
	{
		ADCA.CH0.CTRL |= ADC_CH_START_bm;  // Start conversion
		while(!(ADCA.CH0.INTFLAGS));
		ADCA.CH0.INTFLAGS=ADC_CH_CHIF_bm;
		degrees_x10 += ADCA.CH0RES;
	}	

	degrees_x10 = degrees_x10 / 10;
	degrees_x10 *= kelvin_per_adc_x10;
	degrees_x10 -= 2730;
	return degrees_x10 / 1000;
}

//////////////////////////////////////////////////////////////////////////
/* ADC 

0: PA0: PWR_BACKUP_SIG -   
1: PA1: PWR_CAR_SIG -    ,   
2: PA2: ADC_IN_1 -  
3: PA3: ADC_IN_2 -  
4: PA4: ADC_IN_3 -  
*/
float get_ADC(uint8_t _ch)
{
	float Result;
	
	ADCA.CTRLA = ADC_ENABLE_bm;// enable adc
	
	ADCA.CTRLB = ADC_RESOLUTION_12BIT_gc; // 12 bit conversion
	ADCA.REFCTRL = ADC_REFSEL_VCC_gc;
	ADCA.PRESCALER = ADC_PRESCALER_DIV128_gc; // peripheral clk/128 (32MHz/128=250kHz)
	ADCA.CH0.CTRL = ADC_CH_INPUTMODE_SINGLEENDED_gc;	 // single ended
	
	// 
	switch (_ch)
	{
		case 0:
			ADCA.CH0.MUXCTRL = ADC_CH_MUXPOS_PIN0_gc;
			break;
		
		case 1:
			ADCA.CH0.MUXCTRL = ADC_CH_MUXPOS_PIN1_gc;
			break;

		case 2:
			ADCA.CH0.MUXCTRL = ADC_CH_MUXPOS_PIN2_gc;
			break;

		case 3:
			ADCA.CH0.MUXCTRL = ADC_CH_MUXPOS_PIN3_gc;
			break;
		
		case 4:
			ADCA.CH0.MUXCTRL = ADC_CH_MUXPOS_PIN4_gc;
			break;
	}
	//  
	for(uint8_t _x=0; _x<10; _x++)
	{
		ADCA.CH0.CTRL |= ADC_CH_START_bm;  // Start conversion
		while(!(ADCA.CH0.INTFLAGS));
		ADCA.CH0.INTFLAGS=ADC_CH_CHIF_bm;
		Result += ADCA.CH0RES;
	}	
	
	Result = Result / 10;	
	return (Result * 20.7 / 4096 - ADC_GND_OFFSET);
}

//    
uint8_t get_PIN(uint8_t _num)
{
	switch (_num)
	{
	case 1:
		return PORTB_IN&PIN0_bm; //pos1_in
		break;
	case 2:
		return PORTB_IN&PIN1_bm;//pos2_in
		break;
	case 3:
		return !(PORTB_IN&PIN2_bm);//neg1_in
		break;
	case 4:
		return !(PORTB_IN&PIN3_bm);//neg2_in
		break;
	case 5:
		return PORTC_IN&PIN0_bm;//out_1
		break;
	case 6:
		return PORTC_IN&PIN1_bm;//out_2
		break;
	default:
		return 0;
		break;
	}
}

//, 
ISR(PORTA_INT0_vect)
{
	system.counter_int_1++;
}

ISR(PORTA_INT1_vect)
{
	system.counter_int_2++;
}

//  1  2   
uint16_t get_INTs(uint8_t _num)
{
	uint16_t Result;
	
	switch (_num)
	{
	case 1:
		Result = system.counter_int_1;
		system.counter_int_1 = 0;
		break;
	case 2:
		Result = system.counter_int_2;
		system.counter_int_2 = 0;
		break;
	}
	
	return Result;
}